@@ -1,4 +1,4 @@
-# Copyright (c) 2017 The Regents of the University of Michigan
+# Copyright (c) 2018 The Regents of the University of Michigan
 # All rights reserved.
 # This software is licensed under the BSD 3-Clause License.
 """Detection of compute environments.
@@ -25,8 +25,13 @@ from collections import OrderedDict
 from signac.common import config
 from signac.common import six
 from signac.common.six import with_metaclass
-from . import scheduler
-from . import manage
+
+from .scheduling.base import JobStatus
+from .scheduling.slurm import SlurmScheduler
+from .scheduling.torque import TorqueScheduler
+from .scheduling.simple_scheduler import SimpleScheduler
+from .scheduling.fakescheduler import FakeScheduler
+from .util import config as flow_config
 from .errors import SubmitError
 from .errors import NoSchedulerError
 
@@ -36,7 +41,12 @@ else:
     import importlib.machinery
 
 logger = logging.getLogger(__name__)
-logger.addHandler(logging.NullHandler())
+if six.PY2:
+    logger.addHandler(logging.NullHandler())
+
+
+# Global variable can be used to override detected environment
+ENVIRONMENT = None
 
 
 NUM_NODES_WARNING = """Unable to determine the reqired number of nodes (nn) for this submission.
@@ -54,12 +64,6 @@ or adjust the processors per node (--ppn).
 Alternatively, you can also use --force to ignore this warning.
 """
 
-MISSING_ENV_CONF_KEY_MSG = """Your environment is missing the following configuration key: '{key}'
-Please provide the missing information, for example by adding it to your global configuration:
-
-signac config --global set {key} <VALUE>
-"""
-
 
 def setup(py_modules, **attrs):
     """Setup function for environment modules.
@@ -132,6 +136,10 @@ class JobScript(io.StringIO):
         self._env = env
         super(JobScript, self).__init__()
 
+    def __str__(self):
+        self.seek(0)
+        return self.read()
+
     def write(self, s):
         if six.PY2:
             super(JobScript, self).write(unicode(s))  # noqa
@@ -177,6 +185,7 @@ class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
     scheduler_type = None
     hostname_pattern = None
     submit_flags = None
+    template = 'base_script.sh'
 
     @classmethod
     def script(cls, **kwargs):
@@ -192,7 +201,8 @@ class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
         """Determine whether this specific compute environment is present.
 
         The default method for environment detection is trying to match a
-        hostname pattern.
+        hostname pattern or delegate the environment check to the associated
+        scheduler type.
         """
         if cls.hostname_pattern is None:
             if cls.scheduler_type is None:
@@ -230,9 +240,11 @@ class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
             flags.extend(env_flags)
 
         # Hand off the actual submission to the scheduler
-        script.seek(0)
+        if isinstance(script, JobScript):  # api version < 6
+            script = str(script)
+
         if cls.get_scheduler().submit(script, flags=flags, *args, **kwargs):
-            return manage.JobStatus.submitted
+            return JobStatus.submitted
 
     @staticmethod
     def bg(cmd):
@@ -241,10 +253,17 @@ class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
 
     @classmethod
     def add_args(cls, parser):
+        """Add arguments related to this compute environment to an argument parser.
+
+        :param parser:
+            The argument parser to add arguments to.
+        :type parser:
+            :class:`argparse.ArgumentParser`
+        """
         return
 
     @classmethod
-    def get_config_value(cls, key, default=None):
+    def get_config_value(cls, key, default=flow_config._GET_CONFIG_VALUE_NONE):
         """Request a value from the user's configuration.
 
         This method should be used whenever values need to be provided
@@ -263,46 +282,45 @@ class ComputeEnvironment(with_metaclass(ComputeEnvironmentType)):
         :type key: str
         :param default: A default value in case the key cannot be found
             within the user's configuration.
+        :type key: str
         :return: The value or default value.
         :raises SubmitError: If the key is not in the user's configuration
             and no default value is provided.
         """
-        try:
-            return config.load_config()['flow'][cls.__name__][key]
-        except KeyError:
-            if default is None:
-                k = '{}.{}'.format(cls.__name__, key)
-                print(MISSING_ENV_CONF_KEY_MSG.format(key='flow.' + k))
-                raise SubmitError("Missing environment configuration key: '{}'".format(k))
-            else:
-                return default
+        return flow_config.require_config_value(key, ns=cls.__name__, default=default)
 
 
-class UnknownEnvironment(ComputeEnvironment):
+class StandardEnvironment(ComputeEnvironment):
     "This is a default environment, which is always present."
 
     @classmethod
     def is_present(cls):
         return True
 
-    @classmethod
-    def get_scheduler(cls):
-        raise NoSchedulerError("No scheduler defined for unknown environment.")
-
     @classmethod
     def mpi_cmd(cls, cmd, np):
         return 'mpirun -np {np} {cmd}'.format(np=np, cmd=cmd)
 
 
+class UnknownEnvironment(StandardEnvironment):
+    "Deprecated 'standard' environment, replaced by 'StandardEnvironment.'"
+
+    def __init__(self, *args, **kwargs):
+        super(
+            "The 'UnknownEnvironment' class has been replaced by the "
+            "'StandardEnvironment' class.", DeprecationWarning)
+        super(UnknownEnvironment, self).__init__(*args, **kwargs)
+
+
 class TestEnvironment(ComputeEnvironment):
     """This is a test environment.
 
     The test environment will print a mocked submission script
     and submission commands to screen. This enables testing of
     the job submission script generation in environments without
-    an real scheduler.
+    a real scheduler.
     """
-    scheduler_type = scheduler.FakeScheduler
+    scheduler_type = FakeScheduler
 
     @classmethod
     def mpi_cmd(cls, cmd, np):
@@ -316,9 +334,16 @@ class TestEnvironment(ComputeEnvironment):
         return js
 
 
+class SimpleSchedulerEnvironment(ComputeEnvironment):
+    "An environment for the simple-scheduler scheduler."
+    scheduler_type = SimpleScheduler
+    template = 'simple_scheduler.sh'
+
+
 class TorqueEnvironment(ComputeEnvironment):
     "An environment with TORQUE scheduler."
-    scheduler_type = scheduler.TorqueScheduler
+    scheduler_type = TorqueScheduler
+    template = 'torque.sh'
 
 
 class MoabEnvironment(ComputeEnvironment):
@@ -327,33 +352,23 @@ class MoabEnvironment(ComputeEnvironment):
     This class is deprecated and only kept for backwards
     compatibility.
     """
-    scheduler_type = scheduler.TorqueScheduler
+    scheduler_type = TorqueScheduler
 
     def __init__(self, *args, **kwargs):
-        warnings.warn(
-            "The MoabEnvironment has been replaced by the TorqueEnvironment.",
-            DeprecationWarning)
-        super(MoabEnvironment, self).__init__(*args, **kwargs)
+        raise RuntimeError("The MoabEnvironment has been replaced by the TorqueEnvironment.")
 
 
 class SlurmEnvironment(ComputeEnvironment):
-    "An environment with slurm scheduler."
-    scheduler_type = scheduler.SlurmScheduler
+    "An environment with SLURM scheduler."
+    scheduler_type = SlurmScheduler
+    template = 'slurm.sh'
 
 
 class NodesEnvironment(ComputeEnvironment):
+    """A compute environment consisting of multiple compute nodes.
 
-    @classmethod
-    def add_args(cls, parser):
-        super(NodesEnvironment, cls).add_args(parser)
-        parser.add_argument(
-            '--nn',
-            type=int,
-            help="Specify the number of nodes.")
-        parser.add_argument(
-            '--ppn',
-            type=int,
-            help="Specify the number of processors allocated to each node.")
+    Each compute node is assumed to have a specific number of compute units, e.g., CPUs.
+    """
 
     @classmethod
     def calc_num_nodes(cls, np_total, ppn, force=False, **kwargs):
@@ -373,6 +388,15 @@ class NodesEnvironment(ComputeEnvironment):
                 raise SubmitError("Bad node utilization!")
         return nn
 
+    @classmethod
+    def add_args(cls, parser):
+        super(NodesEnvironment, cls).add_args(parser)
+        parser.add_argument(
+            '--nn',
+            type=int,
+            help="Specify the total number of nodes. This value is computed automatically "
+                 "from the 'np' directive otherwise.")
+
 
 class DefaultTorqueEnvironment(NodesEnvironment, TorqueEnvironment):
     "A default environment for environments with TORQUE scheduler."
@@ -403,14 +427,16 @@ class DefaultTorqueEnvironment(NodesEnvironment, TorqueEnvironment):
             help="Do not copy current environment variables into compute node environment.")
 
     @classmethod
-    def script(cls, _id, nn=None, ppn=None, walltime=None, no_copy_env=False, **kwargs):
+    def gen_tasks(cls, js, np_total):
+        """Helper function to generate the number of tasks (for overriding)"""
+        js.writeline('#PBS -l nodes={}'.format(np_total))
+        return js
+
+    @classmethod
+    def script(cls, _id, np_total, walltime=None, no_copy_env=False, **kwargs):
         js = super(DefaultTorqueEnvironment, cls).script()
         js.writeline('#PBS -N {}'.format(_id))
-        if nn is not None:
-            if ppn is None:
-                js.writeline('#PBS -l nodes={}'.format(nn))
-            else:
-                js.writeline('#PBS -l nodes={}:ppn={}'.format(nn, ppn))
+        js = cls.gen_tasks(js)
         if walltime is not None:
             js.writeline('#PBS -l walltime={}'.format(format_timedelta(walltime)))
         if not no_copy_env:
@@ -419,21 +445,24 @@ class DefaultTorqueEnvironment(NodesEnvironment, TorqueEnvironment):
 
 
 class DefaultSlurmEnvironment(NodesEnvironment, SlurmEnvironment):
-    "A default environment for environments with slurm scheduler."
+    "A default environment for environments with SLURM scheduler."
 
     @classmethod
     def mpi_cmd(cls, cmd, np):
         return 'mpirun -np {np} {cmd}'.format(np=np, cmd=cmd)
 
     @classmethod
-    def script(cls, _id, nn=None, ppn=None, walltime=None, **kwargs):
+    def gen_tasks(cls, js, np_total):
+        """Helper function to generate the number of tasks (for overriding)"""
+        js.writeline('#SBATCH --ntasks={}'.format(np_total))
+        return js
+
+    @classmethod
+    def script(cls, _id, np_total, walltime=None, **kwargs):
         js = super(DefaultSlurmEnvironment, cls).script()
         js.writeline('#!/bin/bash')
         js.writeline('#SBATCH --job-name="{}"'.format(_id))
-        if nn is not None:
-            js.writeline('#SBATCH --nodes={}'.format(nn))
-        if ppn is not None:
-            js.writeline('#SBATCH --ntasks-per-node={}'.format(ppn))
+        js = cls.gen_tasks(js, np_total)
         if walltime is not None:
             js.writeline('#SBATCH -t {}'.format(format_timedelta(walltime)))
         return js
@@ -458,10 +487,12 @@ class DefaultSlurmEnvironment(NodesEnvironment, SlurmEnvironment):
 
 
 class CPUEnvironment(ComputeEnvironment):
+    "An environment with CPUs."
     pass
 
 
 class GPUEnvironment(ComputeEnvironment):
+    "An environment with GPUs."
     pass
 
 
@@ -504,24 +535,46 @@ def get_environment(test=False, import_configured=True):
     first EnvironmentClass where the is_present() method returns
     True.
 
-    :param test: Return the TestEnvironment
-    :type tets: bool
-    :returns: The detected environment class.
+    :param test:
+        Return the TestEnvironment
+    :type test:
+        bool
+    :returns:
+        The detected environment class.
     """
     if test:
         return TestEnvironment
     else:
+        # Return a globally specified environment
+        if ENVIRONMENT is not None:
+            return ENVIRONMENT
+
+        # Obtain a list of all registered environments
         env_types = registered_environments(import_configured=import_configured)
         logger.debug(
             "List of registered environments:\n\t{}".format(
                 '\n\t'.join((str(env.__name__) for env in env_types))))
+
+        # Select environment based on environment variable if set.
+        env_from_env_var = os.environ.get('SIGNAC_FLOW_ENVIRONMENT')
+        if env_from_env_var:
+            for env_type in env_types:
+                if env_type.__name__ == env_from_env_var:
+                    return env_type
+            else:
+                raise ValueError("Unknown environment '{}'.".format(env_from_env_var))
+
+        # Select based on DEBUG flag:
         for env_type in env_types:
             if getattr(env_type, 'DEBUG', False):
                 logger.debug("Select environment '{}'; DEBUG=True.".format(env_type.__name__))
                 return env_type
+
+        # Default selection:
         for env_type in reversed(env_types):
             if env_type.is_present():
                 logger.debug("Select environment '{}'; is present.".format(env_type.__name__))
                 return env_type
-        else:
-            return UnknownEnvironment
+
+        # Otherwise, just return a standard environment
+        return StandardEnvironment
